/*
 * Sweep, a sound wave editor.
 *
 * Copyright (C) 2000 Conrad Parker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <sys/ioctl.h>
#include <pthread.h>

#include <sweep/sweep_types.h>
#include <sweep/sweep_sample.h>

#include "driver.h"

#ifdef DRIVER_OSS

#include <sys/soundcard.h>
#define DEV_DSP "/dev/dsp"

#define NFRAGS 64

typedef struct _oss_play_offset oss_play_offset;

struct _oss_play_offset {
  int framenr;
  sw_framecount_t offset;
};

static oss_play_offset offsets[NFRAGS];

static int oindex;
static int current_frame;
static int frame;

static sw_handle *
open_dev_dsp (void)
{
  int dev_dsp;
  sw_handle * handle;
  int i;

  if((dev_dsp = open(DEV_DSP, O_WRONLY, 0)) == -1) {
    perror ("sweep: unable to open device " DEV_DSP);
    return NULL;
  }

  handle = g_malloc0 (sizeof (sw_handle));
  handle->driver_fd = dev_dsp;

  oindex = 0;
  current_frame = 0;
  for (i = 0; i < NFRAGS; i++) {
    offsets[i].framenr = 0;
    offsets[i].offset = -1;
  }
  frame = 0;

  return handle;
}

static void
setup_dev_dsp (sw_handle * handle, sw_sample * s)
{
  int dev_dsp = handle->driver_fd;
  /*  int mask, format, stereo, frequency;*/

  int stereo = 0;
  int bits;
  int channels;
  int srate;
  int error;
  int numfrags, fragsize, frag;

  if (ioctl (dev_dsp, SNDCTL_DSP_STEREO, &stereo) == -1) {
    /* Fatal error */
    perror("open_dsp_device 2 ") ;
    exit (1);
  } ;

  if (ioctl (dev_dsp, SNDCTL_DSP_RESET, 0)) {
    perror ("open_dsp_device 3 ") ;
    exit (1) ;
  } ;

  numfrags = NFRAGS;
  fragsize = 8;
  frag = (numfrags << 16) | fragsize;
  if ((error = ioctl (dev_dsp, SNDCTL_DSP_SETFRAGMENT, &frag)) != 0) {
    perror ("OSS: error setting fragments");
  }

#ifdef DEBUG
  fragsize = (frag & 0xffff);
  numfrags = (frag & 0x7fff000)>>16;
  g_print ("Got %d frags of size 2^%d\n", numfrags, fragsize);
#endif
    
  bits = 16 ;
  if ((error = ioctl (dev_dsp, SOUND_PCM_WRITE_BITS, &bits)) != 0) {
    perror ("open_dsp_device 4 ") ;
    exit (1) ;
  } ;

  channels = s->sounddata->format->channels;

  if ((error = ioctl (dev_dsp, SOUND_PCM_WRITE_CHANNELS, &channels)) != 0) {
    perror ("open_dsp_device 5 ") ;
    exit (1) ;
  } ;

  handle->driver_channels = channels;

  srate = s->sounddata->format->rate;

  if ((error = ioctl (dev_dsp, SOUND_PCM_WRITE_RATE, &srate)) != 0) {
    perror ("open_dsp_device 6 ") ;
    exit (1) ;
  } ;

  if ((error = ioctl (dev_dsp, SNDCTL_DSP_SYNC, 0)) != 0) {
    perror ("open_dsp_device 7 ") ;
    exit (1) ;
  }

#ifdef DEBUG
  {
    int caps;

    if (ioctl (dev_dsp, SNDCTL_DSP_GETCAPS, &caps) == -1) {
      perror ("OSS: Unable to get device capabilities");
    }
    /* CAP_REALTIME tells whether or not this device can give exact
     * DMA pointers via GETOSPACE/GETISPACE. If this is true, then
     * the device reports with byte accuracy. If it is false it reports
     * to at least the nearest fragment bound, which is still pretty
     * good for small fragments, so it's not much of a problem if
     * this capability is not present.
     */
    g_print ("Realtime: %s\n", caps & DSP_CAP_REALTIME ? "YES" : "NO");
  }
#endif
}

static ssize_t
write_dev_dsp (sw_handle * handle, void * buf, size_t count,
	       sw_framecount_t play_offset)
{
  current_frame += count;
  offsets[oindex].framenr = current_frame;
  offsets[oindex].offset = play_offset;
  oindex++; oindex %= NFRAGS;

  return write (handle->driver_fd, buf, count);
}

static sw_framecount_t
offset_dev_dsp (sw_handle * handle)
{
  count_info info;
  int i, o;

  if (ioctl (handle->driver_fd, SNDCTL_DSP_GETOPTR, &info) == -1) {
    g_print ("error in GETOPTR\n");
    return -1;
  }

  frame = info.bytes;
#ifdef DEBUG
  g_print ("frame: %d\n", frame);
#endif

  o = oindex+1;
  for (i = 0; i < NFRAGS; i++) {
    o %= NFRAGS;
#ifdef DEBUG
    g_print ("\t(%d) Checking %d: %d\n", frame, o, offsets[o].framenr);
#endif
    if (offsets[o].framenr >= frame) {
      return offsets[o].offset;
    }
    o++;
  }

  return -1;
}

static void
reset_dev_dsp (sw_handle * handle)
{
  if(ioctl (handle->driver_fd, SNDCTL_DSP_RESET) == -1) {
    perror ("OSS: error resetting " DEV_DSP);
  }
}

static void
flush_dev_dsp (sw_handle * handle)
{
}

void
drain_dev_dsp (sw_handle * handle)
{
  if(ioctl (handle->driver_fd, SNDCTL_DSP_POST) == -1) {
    perror ("OSS: POST error on " DEV_DSP);
  }

  if (ioctl (handle->driver_fd, SNDCTL_DSP_SYNC, 0) == -1) {
    perror ("OSS: SYNC error on " DEV_DSP) ;
  } ;
}

static void
close_dev_dsp (sw_handle * handle)
{
  close (handle->driver_fd);
}

static sw_driver _driver_oss = {
  open_dev_dsp,
  setup_dev_dsp,
  write_dev_dsp,
  offset_dev_dsp,
  reset_dev_dsp,
  flush_dev_dsp,
  drain_dev_dsp,
  close_dev_dsp,
};

#else

static sw_driver _driver_oss = {
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
};

#endif

sw_driver * driver_oss = &_driver_oss;
